@@ -234,7 +234,7 @@ void parse_linker_script(Context<E> &ctx, MappedFile<Context<E>> *mf) {
 }
 
 template <typename E>
-i64 get_script_output_type(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+MachineType get_script_output_type(Context<E> &ctx, MappedFile<Context<E>> *mf) {
   current_file<E> = mf;
 
   std::vector<std::string_view> vec = tokenize(ctx, mf->get_contents());
@@ -242,11 +242,11 @@ i64 get_script_output_type(Context<E> &ctx, MappedFile<Context<E>> *mf) {
 
   if (tok.size() >= 3 && tok[0] == "OUTPUT_FORMAT" && tok[1] == "(") {
     if (tok[2] == "elf64-x86-64")
-      return EM_X86_64;
+      return MachineType::X86_64;
     if (tok[2] == "elf32-i386")
-      return EM_386;
+      return MachineType::I386;
   }
-  return -1;
+  return MachineType::NONE;
 }
 
 static bool read_label(std::span<std::string_view> &tok,
@@ -395,15 +395,15 @@ void parse_dynamic_list(Context<E> &ctx, std::string path) {
     SyntaxError(ctx, tok[0]) << "trailing garbage token";
 }
 
-#define INSTANTIATE(E)                                                  \
-  template                                                              \
-  void parse_linker_script(Context<E> &ctx, MappedFile<Context<E>> *mf); \
-  template                                                              \
-  i64 get_script_output_type(Context<E> &ctx, MappedFile<Context<E>> *mf); \
-  template                                                              \
-  void parse_version_script(Context<E> &ctx, std::string path);         \
-  template                                                              \
-  void parse_dynamic_list(Context<E> &ctx, std::string path)
+#define INSTANTIATE(E)                                                          \
+  template                                                                      \
+  void parse_linker_script(Context<E> &, MappedFile<Context<E>> *);             \
+  template                                                                      \
+  MachineType get_script_output_type(Context<E> &, MappedFile<Context<E>> *);   \
+  template                                                                      \
+  void parse_version_script(Context<E> &, std::string);                         \
+  template                                                                      \
+  void parse_dynamic_list(Context<E> &, std::string);
 
 INSTANTIATE_ALL;
 
