@@ -2,12 +2,30 @@
 
 namespace mold::macho {
 
+template <typename E>
+OutputSection<E> &get_output_section(Context<E> &ctx, const MachSection &hdr) {
+  static std::unordered_set<std::string_view> set = {
+    "__got", "__auth_got", "__auth_ptr", "__nl_symbol_ptr", "__const",
+    "__cfstring", "__mod_init_func", "__mod_term_func", "__objc_classlist",
+    "__objc_nlclslist", "__objc_catlist", "__objc_nlcatlist", "__objc_protolist",
+  };
+
+  std::string_view seg = hdr.get_segname();
+  std::string_view sect = hdr.get_sectname();
+
+  if (seg == "__DATA" && set.contains(sect)) {
+    seg = "__DATA_CONST";
+  } else if (seg == "__TEXT" && sect == "__StaticInit") {
+    sect = "__text";
+  }
+
+  return *OutputSection<E>::get_instance(ctx, seg, sect);
+}
+
 template <typename E>
 InputSection<E>::InputSection(Context<E> &ctx, ObjectFile<E> &file,
                               const MachSection &hdr)
-  : file(file), hdr(hdr),
-    osec(*OutputSection<E>::get_instance(ctx, hdr.get_segname(),
-                                         hdr.get_sectname())) {
+  : file(file), hdr(hdr), osec(get_output_section(ctx, hdr)) {
   if (hdr.type != S_ZEROFILL)
     contents = file.mf->get_contents().substr(hdr.offset, hdr.size);
 }
