@@ -20,8 +20,8 @@ typedef int16_t i16;
 typedef int32_t i32;
 typedef int64_t i64;
 
-class ARM64;
-class X86_64;
+struct ARM64;
+struct X86_64;
 
 template <typename E>
 std::string rel_to_string(u8 r_type);
@@ -162,13 +162,13 @@ static constexpr u32 S_THREAD_LOCAL_VARIABLE_POINTERS = 0x14;
 static constexpr u32 S_THREAD_LOCAL_INIT_FUNCTION_POINTERS = 0x15;
 static constexpr u32 S_INIT_FUNC_OFFSETS = 0x16;
 
-static constexpr u32 S_ATTR_LOC_RELOC = 0x000001;
-static constexpr u32 S_ATTR_EXT_RELOC = 0x000002;
-static constexpr u32 S_ATTR_SOME_INSTRUCTIONS = 0x000004;
+static constexpr u32 S_ATTR_LOC_RELOC = 0x1;
+static constexpr u32 S_ATTR_EXT_RELOC = 0x2;
+static constexpr u32 S_ATTR_SOME_INSTRUCTIONS = 0x4;
 
-static constexpr u32 S_ATTR_DEBUG = 0x020000;
-static constexpr u32 S_ATTR_SELF_MODIFYING_CODE = 0x040000;
-static constexpr u32 S_ATTR_LIVE_SUPPORT = 0x080000;
+static constexpr u32 S_ATTR_DEBUG = 0x20000;
+static constexpr u32 S_ATTR_SELF_MODIFYING_CODE = 0x40000;
+static constexpr u32 S_ATTR_LIVE_SUPPORT = 0x80000;
 static constexpr u32 S_ATTR_NO_DEAD_STRIP = 0x100000;
 static constexpr u32 S_ATTR_STRIP_STATIC_SYMS = 0x200000;
 static constexpr u32 S_ATTR_NO_TOC = 0x400000;
@@ -200,10 +200,10 @@ static constexpr u32 BIND_TYPE_POINTER = 1;
 static constexpr u32 BIND_TYPE_TEXT_ABSOLUTE32 = 2;
 static constexpr u32 BIND_TYPE_TEXT_PCREL32 = 3;
 
-static constexpr u32 BIND_SPECIAL_DYLIB_SELF = 0;
-static constexpr u32 BIND_SPECIAL_DYLIB_MAIN_EXECUTABLE = -1;
-static constexpr u32 BIND_SPECIAL_DYLIB_FLAT_LOOKUP = -2;
-static constexpr u32 BIND_SPECIAL_DYLIB_WEAK_LOOKUP = -3;
+static constexpr i32 BIND_SPECIAL_DYLIB_SELF = 0;
+static constexpr i32 BIND_SPECIAL_DYLIB_MAIN_EXECUTABLE = -1;
+static constexpr i32 BIND_SPECIAL_DYLIB_FLAT_LOOKUP = -2;
+static constexpr i32 BIND_SPECIAL_DYLIB_WEAK_LOOKUP = -3;
 
 static constexpr u32 BIND_SYMBOL_FLAGS_WEAK_IMPORT = 0x1;
 static constexpr u32 BIND_SYMBOL_FLAGS_NON_WEAK_DEFINITION = 0x8;
@@ -317,6 +317,23 @@ static constexpr u32 PLATFORM_DRIVERKIT = 10;
 static constexpr u32 TOOL_CLANG = 1;
 static constexpr u32 TOOL_SWIFT = 2;
 static constexpr u32 TOOL_LD = 3;
+static constexpr u32 TOOL_MOLD = 54321; // Randomly chosen!
+
+static constexpr u32 OBJC_IMAGE_SUPPORTS_GC = 1 << 1;
+static constexpr u32 OBJC_IMAGE_REQUIRES_GC = 1 << 2;
+static constexpr u32 OBJC_IMAGE_OPTIMIZED_BY_DYLD = 1 << 3;
+static constexpr u32 OBJC_IMAGE_SUPPORTS_COMPACTION = 1 << 4;
+static constexpr u32 OBJC_IMAGE_IS_SIMULATED = 1 << 5;
+static constexpr u32 OBJC_IMAGE_HAS_CATEGORY_CLASS_PROPERTIES = 1 << 6;
+
+static constexpr u32 LOH_ARM64_ADRP_ADRP = 1;
+static constexpr u32 LOH_ARM64_ADRP_LDR = 2;
+static constexpr u32 LOH_ARM64_ADRP_ADD_LDR = 3;
+static constexpr u32 LOH_ARM64_ADRP_LDR_GOT_LDR = 4;
+static constexpr u32 LOH_ARM64_ADRP_ADD_STR = 5;
+static constexpr u32 LOH_ARM64_ADRP_LDR_GOT_STR = 6;
+static constexpr u32 LOH_ARM64_ADRP_ADD = 7;
+static constexpr u32 LOH_ARM64_ADRP_LDR_GOT = 8;
 
 static constexpr u32 ARM64_RELOC_UNSIGNED = 0;
 static constexpr u32 ARM64_RELOC_SUBTRACTOR = 1;
@@ -599,18 +616,27 @@ struct MachSym {
   }
 
   ul32 stroff;
-  u8 is_extern : 1;
-  u8 type : 3;
-  u8 is_private_extern : 1;
-  u8 stub : 3;
+
+  union {
+    u8 n_type;
+    struct {
+      u8 is_extern : 1;
+      u8 type : 3;
+      u8 is_private_extern : 1;
+      u8 stab : 3;
+    };
+  };
+
   u8 sect;
+
   union {
     ul16 desc;
     struct {
       u8 padding;
-      u8 p2align : 4;
+      u8 common_p2align : 4;
     };
   };
+
   ul64 value;
 };
 
@@ -722,9 +748,18 @@ struct CodeSignatureDirectory {
   ub64 exec_seg_flags;
 };
 
+// __DATA,__objc_imageinfo
+struct ObjcImageInfo {
+  ul32 version = 0;
+  u8 flags = 0;
+  u8 swift_version = 0;
+  ul16 swift_lang_version = 0;
+};
+
 struct ARM64 {
   static constexpr u32 cputype = CPU_TYPE_ARM64;
   static constexpr u32 cpusubtype = CPU_SUBTYPE_ARM64_ALL;
+  static constexpr u32 page_size = 16384;
   static constexpr u32 abs_rel = ARM64_RELOC_UNSIGNED;
   static constexpr u32 word_size = 8;
   static constexpr u32 stub_size = 12;
@@ -735,6 +770,7 @@ struct ARM64 {
 struct X86_64 {
   static constexpr u32 cputype = CPU_TYPE_X86_64;
   static constexpr u32 cpusubtype = CPU_SUBTYPE_X86_64_ALL;
+  static constexpr u32 page_size = 4096;
   static constexpr u32 abs_rel = X86_64_RELOC_UNSIGNED;
   static constexpr u32 word_size = 8;
   static constexpr u32 stub_size = 6;
