@@ -3,6 +3,7 @@
 
 #include <shared_mutex>
 #include <sys/mman.h>
+#include <tbb/enumerable_thread_specific.h>
 #include <tbb/parallel_for.h>
 #include <tbb/parallel_for_each.h>
 #include <tbb/parallel_sort.h>
@@ -120,8 +121,8 @@ static std::vector<u8> create_build_version_cmd(Context<E> &ctx) {
   cmd.ntools = 1;
 
   BuildToolVersion &tool = *(BuildToolVersion *)(buf.data() + sizeof(cmd));
-  tool.tool = 3;
-  tool.version = 0x28a0900;
+  tool.tool = TOOL_MOLD;
+  tool.version = parse_version(ctx, MOLD_VERSION);
   return buf;
 }
 
@@ -154,7 +155,13 @@ create_load_dylib_cmd(Context<E> &ctx, DylibFile<E> &dylib) {
   std::vector<u8> buf(align_to(size, 8));
   DylibCommand &cmd = *(DylibCommand *)buf.data();
 
-  cmd.cmd = (dylib.is_weak ? LC_LOAD_WEAK_DYLIB : LC_LOAD_DYLIB);
+  if (dylib.is_reexported)
+    cmd.cmd = LC_REEXPORT_DYLIB;
+  else if (dylib.is_weak)
+    cmd.cmd = LC_LOAD_WEAK_DYLIB;
+  else
+    cmd.cmd = LC_LOAD_DYLIB;
+
   cmd.cmdsize = buf.size();
   cmd.nameoff = sizeof(cmd);
   cmd.timestamp = 2;
@@ -184,8 +191,8 @@ static std::vector<u8> create_function_starts_cmd(Context<E> &ctx) {
 
   cmd.cmd = LC_FUNCTION_STARTS;
   cmd.cmdsize = buf.size();
-  cmd.dataoff = ctx.function_starts.hdr.offset;
-  cmd.datasize = ctx.function_starts.hdr.size;
+  cmd.dataoff = ctx.function_starts->hdr.offset;
+  cmd.datasize = ctx.function_starts->hdr.size;
   return buf;
 }
 
@@ -268,10 +275,12 @@ static std::vector<std::vector<u8>> create_load_commands(Context<E> &ctx) {
     vec.push_back(create_uuid_cmd(ctx));
   vec.push_back(create_build_version_cmd(ctx));
   vec.push_back(create_source_version_cmd(ctx));
-  vec.push_back(create_function_starts_cmd(ctx));
+  if (ctx.arg.function_starts)
+    vec.push_back(create_function_starts_cmd(ctx));
 
-  for (DylibFile<E> *dylib : ctx.dylibs)
-    vec.push_back(create_load_dylib_cmd(ctx, *dylib));
+  for (DylibFile<E> *file : ctx.dylibs)
+    if (file->dylib_idx != BIND_SPECIAL_DYLIB_MAIN_EXECUTABLE)
+      vec.push_back(create_load_dylib_cmd(ctx, *file));
 
   for (std::string_view rpath : ctx.arg.rpath)
     vec.push_back(create_rpath_cmd(ctx, rpath));
@@ -313,6 +322,14 @@ static bool has_tlv(Context<E> &ctx) {
   return false;
 }
 
+template <typename E>
+static bool has_reexported_lib(Context<E> &ctx) {
+  for (DylibFile<E> *file : ctx.dylibs)
+    if (file->is_reexported)
+      return true;
+  return false;
+}
+
 template <typename E>
 void OutputMachHeader<E>::copy_buf(Context<E> &ctx) {
   u8 *buf = ctx.buf + this->hdr.offset;
@@ -331,9 +348,12 @@ void OutputMachHeader<E>::copy_buf(Context<E> &ctx) {
   if (has_tlv(ctx))
     mhdr.flags |= MH_HAS_TLV_DESCRIPTORS;
 
-  if (ctx.output_type == MH_DYLIB)
+  if (ctx.output_type == MH_DYLIB && !has_reexported_lib(ctx))
     mhdr.flags |= MH_NO_REEXPORTED_DYLIBS;
 
+  if (ctx.arg.mark_dead_strippable_dylib)
+    mhdr.flags |= MH_DEAD_STRIPPABLE_DYLIB;
+
   write_vector(buf + sizeof(mhdr), flatten(cmds));
 }
 
@@ -651,7 +671,8 @@ void RebaseSection<E>::compute_size(Context<E> &ctx) {
       if (chunk->is_output_section && !chunk->hdr.match("__TEXT", "__eh_frame"))
         for (Subsection<E> *subsec : ((OutputSection<E> *)chunk)->members)
           for (Relocation<E> &rel : subsec->get_rels())
-            if (!rel.is_pcrel && rel.type == E::abs_rel && !refers_tls(rel.sym))
+            if (!rel.is_pcrel && !rel.is_subtracted && rel.type == E::abs_rel &&
+                !refers_tls(rel.sym))
               enc.add(seg->seg_idx,
                       subsec->get_addr(ctx) + rel.offset - seg->cmd.vmaddr);
 
@@ -670,7 +691,7 @@ BindEncoder::BindEncoder() {
 }
 
 template <typename E>
-static u32 get_dylib_idx(InputFile<E> *file) {
+static i32 get_dylib_idx(InputFile<E> *file) {
   if (file->is_dylib)
     return ((DylibFile<E> *)file)->dylib_idx;
   return BIND_SPECIAL_DYLIB_FLAT_LOOKUP;
@@ -679,10 +700,13 @@ static u32 get_dylib_idx(InputFile<E> *file) {
 template <typename E>
 void BindEncoder::add(Symbol<E> &sym, i64 seg_idx, i64 offset) {
   i64 dylib_idx = get_dylib_idx(sym.file);
-  i64 flags = (sym.file->is_weak ? BIND_SYMBOL_FLAGS_WEAK_IMPORT : 0);
+  i64 flags = (sym.is_weak ? BIND_SYMBOL_FLAGS_WEAK_IMPORT : 0);
 
   if (last_dylib != dylib_idx) {
-    if (dylib_idx < 16) {
+    if (dylib_idx < 0) {
+      buf.push_back(BIND_OPCODE_SET_DYLIB_SPECIAL_IMM |
+                    (dylib_idx & BIND_IMMEDIATE_MASK));
+    } else if (dylib_idx < 16) {
       buf.push_back(BIND_OPCODE_SET_DYLIB_ORDINAL_IMM | dylib_idx);
     } else {
       buf.push_back(BIND_OPCODE_SET_DYLIB_ORDINAL_ULEB);
@@ -759,14 +783,16 @@ void LazyBindSection<E>::add(Context<E> &ctx, Symbol<E> &sym) {
 
   i64 dylib_idx = get_dylib_idx(sym.file);
 
-  if (dylib_idx < 16) {
+  if (dylib_idx < 0) {
+    emit(BIND_OPCODE_SET_DYLIB_SPECIAL_IMM | (dylib_idx & BIND_IMMEDIATE_MASK));
+  } else if (dylib_idx < 16) {
     emit(BIND_OPCODE_SET_DYLIB_ORDINAL_IMM | dylib_idx);
   } else {
     emit(BIND_OPCODE_SET_DYLIB_ORDINAL_ULEB);
     encode_uleb(contents, dylib_idx);
   }
 
-  i64 flags = (sym.file->is_weak ? BIND_SYMBOL_FLAGS_WEAK_IMPORT : 0);
+  i64 flags = (sym.is_weak ? BIND_SYMBOL_FLAGS_WEAK_IMPORT : 0);
   assert(flags < 16);
 
   emit(BIND_OPCODE_SET_SYMBOL_TRAILING_FLAGS_IMM | flags);
@@ -832,8 +858,7 @@ i64 ExportEncoder::finish() {
 
 static i64 common_prefix_len(std::string_view x, std::string_view y) {
   i64 i = 0;
-  i64 end = std::min(x.size(), y.size());
-  while (i < end && x[i] == y[i])
+  while (i < x.size() && i < y.size() && x[i] == y[i])
     i++;
   return i;
 }
@@ -864,7 +889,7 @@ ExportEncoder::construct_trie(TrieNode &node, std::span<Entry> entries, i64 len,
     std::span<Entry> subspan = entries.subspan(i, j - i);
 
     if (divide && j - i < grain_size) {
-      tg->run([=] {
+      tg->run([=, this] {
         construct_trie(*child, subspan, new_len, tg, grain_size, false);
       });
     } else {
@@ -925,20 +950,31 @@ template <typename E>
 void ExportSection<E>::compute_size(Context<E> &ctx) {
   for (ObjectFile<E> *file : ctx.objs)
     for (Symbol<E> *sym : file->syms)
-      if (sym && sym->file == file & sym->scope == SCOPE_EXTERN)
-        enc.entries.push_back({sym->name, 0,
-                               sym->get_addr(ctx) - ctx.arg.pagezero_size});
+      if (sym && sym->file == file && sym->scope == SCOPE_EXTERN)
+        enc.entries.push_back({
+            sym->name,
+            sym->is_weak ? EXPORT_SYMBOL_FLAGS_WEAK_DEFINITION : 0,
+            sym->get_addr(ctx) - ctx.arg.pagezero_size});
+
+  if (enc.entries.empty())
+    return;
 
   this->hdr.size = align_to(enc.finish(), 8);
 }
 
 template <typename E>
 void ExportSection<E>::copy_buf(Context<E> &ctx) {
+  if (this->hdr.size == 0)
+    return;
+
   u8 *buf = ctx.buf + this->hdr.offset;
   memset(buf, 0, this->hdr.size);
   enc.write_trie(buf, enc.root);
 }
 
+// LC_FUNCTION_STARTS contains function start addresses encoded in
+// ULEB128. I don't know what tools consume this table, but we create
+// it anyway by default for the sake of compatibility.
 template <typename E>
 void FunctionStartsSection<E>::compute_size(Context<E> &ctx) {
   std::vector<std::vector<u64>> vec(ctx.objs.size());
@@ -976,17 +1012,39 @@ void FunctionStartsSection<E>::copy_buf(Context<E> &ctx) {
 template <typename E>
 void SymtabSection<E>::compute_size(Context<E> &ctx) {
   symtab_offsets.clear();
-  strtab_offsets.clear();
-
   symtab_offsets.resize(ctx.objs.size() + ctx.dylibs.size() + 1);
+
+  strtab_offsets.clear();
   strtab_offsets.resize(ctx.objs.size() + ctx.dylibs.size() + 1);
+  strtab_offsets[0] = 1;
+
+  tbb::enumerable_thread_specific<i64> locals;
+  tbb::enumerable_thread_specific<i64> globals;
+  tbb::enumerable_thread_specific<i64> undefs;
+
+  // Calculate the sizes for -add_ast_path symbols.
+  locals.local() += ctx.arg.add_ast_path.size();
+  symtab_offsets[0] += ctx.arg.add_ast_path.size();
+  for (std::string_view s : ctx.arg.add_ast_path)
+    strtab_offsets[0] += s.size() + 1;
+
+  // Calculate the sizes required for symbols in input files.
+  auto count = [&](Symbol<E> *sym) {
+    if (sym->is_imported)
+      undefs.local() += 1;
+    else if (sym->scope == SCOPE_EXTERN)
+      globals.local() += 1;
+    else
+      locals.local() += 1;
+  };
 
   tbb::parallel_for((i64)0, (i64)ctx.objs.size(), [&](i64 i) {
     ObjectFile<E> &file = *ctx.objs[i];
     for (Symbol<E> *sym : file.syms) {
-      if (sym && sym->file == &file) {
+      if (sym && sym->file == &file && (!sym->subsec || sym->subsec->is_alive)) {
         symtab_offsets[i + 1]++;
         strtab_offsets[i + 1] += sym->name.size() + 1;
+        count(sym);
       }
     }
   });
@@ -998,14 +1056,18 @@ void SymtabSection<E>::compute_size(Context<E> &ctx) {
           (sym->stub_idx != -1 || sym->got_idx != -1)) {
         symtab_offsets[i + 1 + ctx.objs.size()]++;
         strtab_offsets[i + 1 + ctx.objs.size()] += sym->name.size() + 1;
+        count(sym);
       }
     }
   });
 
+  num_locals = locals.combine(std::plus());
+  num_globals = globals.combine(std::plus());
+  num_undefs = undefs.combine(std::plus());
+
   for (i64 i = 1; i < symtab_offsets.size(); i++)
     symtab_offsets[i] += symtab_offsets[i - 1];
 
-  strtab_offsets[0] = 1;
   for (i64 i = 1; i < strtab_offsets.size(); i++)
     strtab_offsets[i] += strtab_offsets[i - 1];
 
@@ -1020,20 +1082,47 @@ void SymtabSection<E>::copy_buf(Context<E> &ctx) {
   u8 *strtab = ctx.buf + ctx.strtab.hdr.offset;
   strtab[0] = '\0';
 
+  // Create symbols for -add_ast_path
+  {
+    i64 symoff = 0;
+    i64 stroff = 1;
+    for (std::string_view s : ctx.arg.add_ast_path) {
+      MachSym &msym = buf[symoff++];
+      msym.stroff = stroff;
+      msym.n_type = N_AST;
+
+      write_string(strtab + stroff, s);
+      stroff += s.size() + 1;
+    }
+  }
+
+  // Copy symbols from input files to an output file
+  Symbol<E> *mh_execute_header = get_symbol(ctx, "__mh_execute_header");
+  Symbol<E> *dyld_private = get_symbol(ctx, "__dyld_private");
+  Symbol<E> *mh_dylib_header = get_symbol(ctx, "__mh_dylib_header");
+  Symbol<E> *mh_bundle_header = get_symbol(ctx, "__mh_bundle_header");
+  Symbol<E> *dso_handle = get_symbol(ctx, "___dso_handle");
+
   auto write = [&](Symbol<E> &sym, i64 symoff, i64 stroff) {
     MachSym &msym = buf[symoff];
 
     msym.stroff = stroff;
     write_string(strtab + stroff, sym.name);
 
-    msym.type = (sym.is_imported ? N_UNDF : N_SECT);
     msym.is_extern = (sym.is_imported || sym.scope == SCOPE_EXTERN);
+    msym.type = (sym.is_imported ? N_UNDF : N_SECT);
 
-    if (!sym.is_imported && (!sym.subsec || sym.subsec->is_alive))
-      msym.value = sym.get_addr(ctx);
-
-    if (sym.subsec && sym.subsec->is_alive)
+    if (sym.is_imported)
+      msym.sect = N_UNDF;
+    else if (sym.subsec)
       msym.sect = sym.subsec->isec.osec.sect_idx;
+    else if (&sym == mh_execute_header)
+      msym.sect = ctx.text->sect_idx;
+    else if (&sym == dyld_private || &sym == mh_dylib_header ||
+             &sym == mh_bundle_header || &sym == dso_handle)
+      msym.sect = ctx.data->sect_idx;
+    else
+      msym.sect = N_ABS;
 
     if (sym.file->is_dylib)
       msym.desc = ((DylibFile<E> *)sym.file)->dylib_idx << 8;
@@ -1041,6 +1130,9 @@ void SymtabSection<E>::copy_buf(Context<E> &ctx) {
       msym.desc = DYNAMIC_LOOKUP_ORDINAL << 8;
     else if (sym.referenced_dynamically)
       msym.desc = REFERENCED_DYNAMICALLY;
+
+    if (!sym.is_imported && (!sym.subsec || sym.subsec->is_alive))
+      msym.value = sym.get_addr(ctx);
   };
 
   tbb::parallel_for((i64)0, (i64)ctx.objs.size(), [&](i64 i) {
@@ -1049,7 +1141,7 @@ void SymtabSection<E>::copy_buf(Context<E> &ctx) {
     i64 stroff = strtab_offsets[i];
 
     for (Symbol<E> *sym : file.syms) {
-      if (sym && sym->file == &file) {
+      if (sym && sym->file == &file && (!sym->subsec || sym->subsec->is_alive)) {
         write(*sym, symoff, stroff);
         symoff++;
         stroff += sym->name.size() + 1;
@@ -1071,21 +1163,91 @@ void SymtabSection<E>::copy_buf(Context<E> &ctx) {
       }
     }
   });
+
+  auto get_rank = [](const MachSym &msym) {
+    if (msym.sect == N_UNDF)
+      return 2;
+    if (msym.is_extern)
+      return 1;
+    return 0;
+  };
+
+  std::stable_sort(buf, buf + this->hdr.size / sizeof(MachSym),
+       [&](const MachSym &a, const MachSym &b) {
+     return std::tuple{get_rank(a), a.value} < std::tuple{get_rank(b), b.value};
+  });
+}
+
+template <typename E>
+static bool has_objc_image_info_section(Context<E> &ctx) {
+  return false;
+}
+
+// Create __DATA,__objc_imageinfo section contents by merging input
+// __objc_imageinfo sections.
+template <typename E>
+std::unique_ptr<ObjcImageInfoSection<E>>
+ObjcImageInfoSection<E>::create(Context<E> &ctx) {
+  ObjcImageInfo *first = nullptr;
+
+  for (ObjectFile<E> *file : ctx.objs) {
+    if (file->objc_image_info) {
+      first = file->objc_image_info;
+      break;
+    }
+  }
+
+  if (!first)
+    return nullptr;
+
+  ObjcImageInfo info;
+  info.flags = (first->flags & OBJC_IMAGE_HAS_CATEGORY_CLASS_PROPERTIES);
+
+  for (ObjectFile<E> *file : ctx.objs) {
+    if (!file->objc_image_info)
+      continue;
+
+    ObjcImageInfo &info2 = *file->objc_image_info;
+
+    // Make sure that all object files have the same flag.
+    if ((info.flags & OBJC_IMAGE_HAS_CATEGORY_CLASS_PROPERTIES) !=
+        (info2.flags & OBJC_IMAGE_HAS_CATEGORY_CLASS_PROPERTIES))
+      Error(ctx) << *file << ": incompatible __objc_imageinfo flag";
+
+    // Make sure that all object files have the same Swift version.
+    if (info.swift_version == 0)
+      info.swift_version = info2.swift_version;
+
+    if (info.swift_version != info2.swift_version && info2.swift_version != 0)
+      Error(ctx) << *file << ": incompatible __objc_imageinfo swift version"
+                 << (u32)info.swift_version << " " << (u32)info2.swift_version;
+
+    // swift_lang_version is set to the newest.
+    info.swift_lang_version =
+      std::max<u32>(info.swift_lang_version, info2.swift_lang_version);
+  }
+
+  return std::make_unique<ObjcImageInfoSection<E>>(ctx, info);
+}
+
+template <typename E>
+void ObjcImageInfoSection<E>::copy_buf(Context<E> &ctx) {
+  memcpy(ctx.buf + this->hdr.offset, &contents, sizeof(contents));
 }
 
 template <typename E>
 void CodeSignatureSection<E>::compute_size(Context<E> &ctx) {
   std::string filename = filepath(ctx.arg.final_output).filename();
   i64 filename_size = align_to(filename.size() + 1, 16);
-  i64 num_blocks = align_to(this->hdr.offset, BLOCK_SIZE) / BLOCK_SIZE;
+  i64 num_blocks = align_to(this->hdr.offset, E::page_size) / E::page_size;
   this->hdr.size = sizeof(CodeSignatureHeader) + sizeof(CodeSignatureBlobIndex) +
                    sizeof(CodeSignatureDirectory) + filename_size +
                    num_blocks * SHA256_SIZE;
 }
 
 // A __code_signature section is optional for x86 macOS but mandatory
-// for ARM macOS. The section contains a cryptographic hash for each 4
-// KiB block of an executable or a dylib file. The program loader
+// for ARM macOS. The section contains a cryptographic hash for each
+// memory page of an executable or a dylib file. The program loader
 // verifies the hash values on the initial execution of a binary and
 // will reject it if a hash value does not match.
 template <typename E>
@@ -1097,7 +1259,7 @@ void CodeSignatureSection<E>::write_signature(Context<E> &ctx) {
 
   std::string filename = filepath(ctx.arg.final_output).filename();
   i64 filename_size = align_to(filename.size() + 1, 16);
-  i64 num_blocks = align_to(this->hdr.offset, BLOCK_SIZE) / BLOCK_SIZE;
+  i64 num_blocks = align_to(this->hdr.offset, E::page_size) / E::page_size;
 
   // Fill code-sign header fields
   CodeSignatureHeader &sighdr = *(CodeSignatureHeader *)buf;
@@ -1126,7 +1288,7 @@ void CodeSignatureSection<E>::write_signature(Context<E> &ctx) {
   dir.code_limit = this->hdr.offset;
   dir.hash_size = SHA256_SIZE;
   dir.hash_type = CS_HASHTYPE_SHA256;
-  dir.page_size = std::countr_zero<u64>(BLOCK_SIZE);
+  dir.page_size = std::countr_zero(E::page_size);
   dir.exec_seg_base = ctx.text_seg->cmd.fileoff;
   dir.exec_seg_limit = ctx.text_seg->cmd.filesize;
   if (ctx.output_type == MH_EXECUTE)
@@ -1135,23 +1297,21 @@ void CodeSignatureSection<E>::write_signature(Context<E> &ctx) {
   memcpy(buf, filename.data(), filename.size());
   buf += filename_size;
 
-  // Compute a hash value for each 4 KiB block. The block size must be
-  // 4 KiB, as the macOS kernel supports only that block size for the
-  // ad-hoc code signatures.
+  // Compute a hash value for each block.
   auto compute_hash = [&](i64 i) {
-    u8 *start = ctx.buf + i * BLOCK_SIZE;
-    u8 *end = ctx.buf + std::min<i64>((i + 1) * BLOCK_SIZE, this->hdr.offset);
+    u8 *start = ctx.buf + i * E::page_size;
+    u8 *end = ctx.buf + std::min<i64>((i + 1) * E::page_size, this->hdr.offset);
     SHA256(start, end - start, buf + i * SHA256_SIZE);
   };
 
   for (i64 i = 0; i < num_blocks; i += 1024) {
     i64 j = std::min(num_blocks, i + 1024);
-    tbb::parallel_for(i, j, [&](i64 k) { compute_hash(k); });
+    tbb::parallel_for(i, j, compute_hash);
 
 #if __APPLE__
     // Calling msync() with MS_ASYNC speeds up the following msync()
     // with MS_INVALIDATE.
-    msync(ctx.buf + i * BLOCK_SIZE, 1024 * BLOCK_SIZE, MS_ASYNC);
+    msync(ctx.buf + i * E::page_size, 1024 * E::page_size, MS_ASYNC);
 #endif
   }
 
@@ -1159,7 +1319,7 @@ void CodeSignatureSection<E>::write_signature(Context<E> &ctx) {
   // entire file. We compute its value as a tree hash.
   if (ctx.arg.uuid == UUID_HASH) {
     u8 uuid[SHA256_SIZE];
-    SHA256(buf, num_blocks * SHA256_SIZE, uuid);
+    SHA256(ctx.buf + this->hdr.offset, this->hdr.size, uuid);
 
     // Indicate that this is UUIDv4 as defined by RFC4122.
     uuid[6] = (uuid[6] & 0b00001111) | 0b01010000;
@@ -1171,7 +1331,7 @@ void CodeSignatureSection<E>::write_signature(Context<E> &ctx) {
     // recompute code signatures for the updated blocks.
     ctx.mach_hdr.copy_buf(ctx);
 
-    for (i64 i = 0; i * BLOCK_SIZE < ctx.mach_hdr.hdr.size; i++)
+    for (i64 i = 0; i * E::page_size < ctx.mach_hdr.hdr.size; i++)
       compute_hash(i);
   }
 
@@ -1332,7 +1492,6 @@ UnwindEncoder<E>::encode(Context<E> &ctx, std::span<UnwindRecord<E>> records) {
 
     page1++;
     page2 = (UnwindSecondLevelPage *)(encoding + map.size());
-    break;
   }
 
   // Write a terminator
@@ -1457,6 +1616,15 @@ void ThreadPtrsSection<E>::copy_buf(Context<E> &ctx) {
       buf[i] = sym.get_addr(ctx);
 }
 
+template <typename E>
+SectCreateSection<E>::SectCreateSection(Context<E> &ctx, std::string_view seg,
+                                        std::string_view sect,
+                                        std::string_view contents)
+  : Chunk<E>(ctx, seg, sect), contents(contents) {
+  this->hdr.size = contents.size();
+  ctx.chunk_pool.emplace_back(this);
+}
+
 template <typename E>
 void SectCreateSection<E>::copy_buf(Context<E> &ctx) {
   write_string(ctx.buf + this->hdr.offset, contents);
@@ -1474,6 +1642,7 @@ void SectCreateSection<E>::copy_buf(Context<E> &ctx) {
   template class SymtabSection<E>;                      \
   template class StrtabSection<E>;                      \
   template class CodeSignatureSection<E>;               \
+  template class ObjcImageInfoSection<E>;               \
   template class DataInCodeSection<E>;                  \
   template class StubsSection<E>;                       \
   template class StubHelperSection<E>;                  \
