@@ -9,8 +9,7 @@ OBJDUMP="${OBJDUMP:-objdump}"
 MACHINE="${MACHINE:-$(uname -m)}"
 testname=$(basename "$0" .sh)
 echo -n "Testing $testname ... "
-cd "$(dirname "$0")"/../..
-t=out/test/elf/$testname
+t=out/test/elf/$MACHINE/$testname
 mkdir -p $t
 
 [ $MACHINE = $(uname -m) ] || { echo skipped; exit; }
@@ -53,8 +52,8 @@ $QEMU $t/exe | grep -q 'Hello world'
 DEBUGINFOD_URLS= gdb $t/exe -nx -batch -ex 'b main' -ex r -ex 'b trap' \
   -ex c -ex bt -ex quit >& $t/log
 
-grep -Pq 'hello \(\) at .*<stdin>:7' $t/log
-grep -Pq 'greet \(\) at .*<stdin>:11' $t/log
-grep -Pq 'main \(\) at .*<stdin>:4' $t/log
+grep -q 'hello () at .*<stdin>:7' $t/log
+grep -q 'greet () at .*<stdin>:11' $t/log
+grep -q 'main () at .*<stdin>:4' $t/log
 
 echo OK
