@@ -9,13 +9,14 @@ OBJDUMP="${OBJDUMP:-objdump}"
 MACHINE="${MACHINE:-$(uname -m)}"
 testname=$(basename "$0" .sh)
 echo -n "Testing $testname ... "
-cd "$(dirname "$0")"/../..
-t=out/test/elf/$testname
+t=out/test/elf/$MACHINE/$testname
 mkdir -p $t
 
+# Skip if libc is musl because musl does not support GNU FUNC
+ldd --help 2>&1 | grep -q musl && { echo skipped; exit; }
+
 # We need to implement R_386_GOT32X relaxation to support PIE on i386
-[ $MACHINE = i386 ] && { echo skipped; exit; }
-[ $MACHINE = i686 ] && { echo skipped; exit; }
+[ $MACHINE = i386 -o $MACHINE = i686 ] && { echo skipped; exit; }
 
 # RISCV64 does not support IFUNC yet
 [ $MACHINE = riscv64 ] && { echo skipped; exit; }
