@@ -9,14 +9,13 @@ OBJDUMP="${OBJDUMP:-objdump}"
 MACHINE="${MACHINE:-$(uname -m)}"
 testname=$(basename "$0" .sh)
 echo -n "Testing $testname ... "
-cd "$(dirname "$0")"/../..
-t=out/test/elf/$testname
+t=out/test/elf/$MACHINE/$testname
 mkdir -p $t
 
 # Skip if target is not x86-64
 [ $MACHINE = x86_64 ] || { echo skipped; exit; }
 
-$CC -fcf-protection=branch /dev/null -o /dev/null -xc - 2> /dev/null || \
+$CC -fcf-protection=branch /dev/null -o /dev/null -xc 2> /dev/null || \
   { echo skipped; exit; }
 
 cat <<EOF | $CC -fcf-protection=branch -c -o $t/a.o -xc -
