@@ -9,8 +9,7 @@ OBJDUMP="${OBJDUMP:-objdump}"
 MACHINE="${MACHINE:-$(uname -m)}"
 testname=$(basename "$0" .sh)
 echo -n "Testing $testname ... "
-cd "$(dirname "$0")"/../..
-t=out/test/elf/$testname
+t=out/test/elf/$MACHINE/$testname
 mkdir -p $t
 
 [ $MACHINE = x86_64 ] || { echo skipped; exit; }
@@ -36,6 +35,6 @@ EOF
 
 $OBJDUMP -d -j .plt.got $t/exe > $t/log
 
-grep -Eq '1020:.*jmp.* <ext2>' $t/log
+grep -Eq '1034:.*jmp.* <ext2>' $t/log
 
 echo OK
