@@ -413,7 +413,7 @@ namespace r1 {
         int flags = RTLD_NOW;
         if (local_binding) {
             flags = flags | RTLD_LOCAL;
-#if __linux__ && !__ANDROID__ && !__TBB_USE_SANITIZERS
+#if __linux__ && __GLIBC__ && !__TBB_USE_SANITIZERS
             flags = flags | RTLD_DEEPBIND;
 #endif
         } else {
